/*
 * This file is part of John the Ripper password cracker,
 * Copyright (c) 1996-2001 by Solar Designer
 *
 * 'Plugin' module for 1.7.8 jumbo patch.
 * Converted to OpenVMS format module by David Jones
 *
 * Date:  9-JUL-2009
 * Revised: 25-JUL-2009		Replace long long references with uaf_qword.
 * Revised: 26-AUG-2011		Update for 1.7.8, move pthread.h include.
 *
 * Copyright (c) 2011 by David L. Jones <jonesd/at/columbus.rr.com>, and
 * is hereby released to the general public under the following terms:
 *    Redistribution and use in source and binary forms, with or without
 *    modifications, are permitted.
 */
#include <stdio.h>
#include <string.h>

#include "arch.h"
#include "misc.h"

#include "common.h"
#include "formats.h"
#include "logger.h"

/*
 * OpenVMS Purdy-based password hash implementation.
 */

#ifndef _JOHN_VMS_STD_H
#define _JOHN_VMS_STD_H

typedef ARCH_WORD_32 VMS_word;

/*
 * Define functions to encode/deocde SYSUAF password information as a
 * printable string suitable for use in a UNIX-style passwd file (i.e. no
 * colons)
 *
 * Copyright (c) 2011 by David L. Jones <jonesd/at/columbus.rr.com>, and
 * is hereby released to the general public under the following terms:
 *    Redistribution and use in source and binary forms, with or without
 *    modifications, are permitted.
 */
#ifndef _JOHN_UAF_ENCODE_H
#define _JOHN_UAF_ENCODE_H
#define UAF_ENCODE_SIZE 30
#include <limits.h>
#if UINT_MAX == ULONG_MAX
typedef unsigned long uaf_lword;
#else
#if UINT_MAX >= 2147483648
typedef unsigned int uaf_lword;	/* sizeof(long)==sizeof(long long)? */
#else
typedef unsigned long uaf_lword;/* int < 32 bits, too short */
#endif
#endif
struct uaf_packed_text {
	unsigned short  r40[4];	/* 12 chars, 3/word */
	char            s[14];	/* Expanded to 1 char/byte */
};
#ifdef NOLONGLONG
#define HASH_INDEX(result,mask) result.info.hash.lw[LSL]&mask)
#else
#define HASH_INDEX(result,mask) result.info.hash.&mask)
#endif

#ifdef NOLONGLONG
#define MSL 1
#define LSL 0			/* little endian ordering */
typedef struct {
	unsigned long   lw[2];
}               uaf_qword;
#define UAF_QW_EQL(a,b) (((a).lw[0]==(b).lw[0]) && ((a).lw[1]==(b).lw[1]))
#define UAF_QW_GEQ(a,b) ((a).lw[MSL]>(b).lw[MSL]) || \
    ( ((a).lw[MSL]==(b).lw[MSL]) && ((a).lw[LSL]>=(b).lw[LSL]) )
#define UAF_QW_SET(a,n) (a).lw[LSL]=(n); (a).lw[MSL]=0
#define UAF_QW_ADD(q,l) { q.lw[LSL]+=l; if (q.lw[LSL] < l) q.lw[MSL]++; }
#define UAF_QW_AND(q,lwmask) ((q).lw[LSL]&lwmask)
#else
typedef unsigned long long uaf_qword;
#define UAF_QW_EQL(a,b) (a==b)
#define UAF_QW_GEQ(a,b) (a>=b)
#define UAF_QW_SET(a,n) a=n
#define UAF_QW_ZEXT(a,n) a=n
#define UAF_QW_ADD(q,l) q += (l);
#define UAF_QW_AND(q,lwmask) ((q)&lwmask)
#endif
struct uaf_hash_info {
	uaf_qword       hash;	/* UAI$_PWd */
	long            flags;	/* UAI$_FLAGS, only bits preseved: */
	/* UAI$M_PWDMIX  */
	/* UAI$M_EXTAUTH */
	unsigned short  salt;	/* UAI$_SALT */
	unsigned char   alg;	/* UAI$_ALGORITHM */
	unsigned char   opt;	/* <0> = (flags&UAI$M_PWDMIX) */
	struct uaf_packed_text username;	/* RAD-50 encoded username */
};
/*
 * Structure is filled in with zero-terminated strings.
 */
struct uaf_account_info {
	char            username[32];	/* Username, upcased */
	unsigned short  uic[2];
	char            owner[32];	/* Name of account owner */
	char            home_dir[96];
	char            shell[32];
};
/*
 * Basic functions for converting UAF information between binary form and
 * printable string (form used to store in passwd file).
 */
char           *
uaf_hash_encode(
	struct uaf_hash_info * pwd,	/* Input argument */
	char encoded[UAF_ENCODE_SIZE]);	/* Output buffer */

	int             uaf_hash_decode(
	                char *encoded,	/* Input argument */
	                struct uaf_hash_info * pwd);	/* Output buffer */
	int             uaf_packed_convert(struct uaf_packed_text * username, int to_packed);
/*
 * Function to hash a clear text password with the salt value in a uaf_hash_info
 * structure and compare the result with the hash value stored there.
 */
	int             uaf_test_password(
	                struct uaf_hash_info * pwd,
	                const char *password,	/* clear text password */
	                int replace_if);	/* Update pwd if false */

	int             uaf_init(void);	/* one-time init for hash func. */
/*
 * Wapper for system function to retrieve authorization information for SYSUAF
 * data file /
 * compute hash from password.  Return value is VMS status code.
 */
	int             uaf_getuai_info(
	                const char *username,	/* Username to find */
	                struct uaf_hash_info * pwd,	/* Password info from
							 * UAF record */
	                struct uaf_hash_info * pwd2,	/* secondary password
							 * info */
	                struct uaf_account_info * acct);

#endif

/*
 * Various structures for internal use.
 */

#define VMS_ALGORITHM_NAME		"32/" ARCH_BITS_STR
/*
 * Global variables shared between vms_std and vms_fmt.  HINT_GENERATION_MASK
 * defines portion of seen[] value that holds generation number, remainder
 * of bits <HINT_GENMASK_BITS:31> hold index of matching password or 0xff if
 * multiple matches.
 */
	extern int      VMS_dbg_flag;
#ifndef SAVE_LIMIT
#define SAVE_LIMIT 240		/* divisible by
				 * 2,3,4,5,6,8,12,15,20,30,40,60,80 */
#define HINT_GENMASK_BITS 24
#define HINT_GENERATION_MASK 0x0ffffff	/* field to hold generation # */
#define HINT_MULTIPLE 0x0ff	/* must be > SAVE_LIMIT */
#endif
	struct result_hint {
		void           *lock_vp;	/* serialization object */
		uaf_lword       seq;	/* crypt_all generation number */
		uaf_lword       nxtseq;
		uaf_lword       mask;	/* must be power of 2 minus 1 */
		uaf_lword       seen[1];	/* variably sized, power of 2 */
	};
	extern struct result_hint *VMS_std_hint;

	struct result_hash {
		struct uaf_hash_info info;
		char            cache_line[64 - sizeof(struct uaf_hash_info)];
	};
	extern struct result_hash *VMS_std_crypt_out;
/*
 * Initializes the internal structures.
 */
	extern void     VMS_std_init(struct fmt_main *);
/*
 * Sets a salt for VMS_std_crypt().
 */
	extern void     VMS_std_set_salt(void *salt);

/*
 * Sets a key for VMS_std_crypt().
 * Currently only supports keys up to 15 characters long.
 */
	extern int      VMS_std_set_key(char *key, int index);
	extern char    *VMS_std_get_key(int position, int index);

/*
 * Main encryption routine, sets VMS_out.
 */
	extern void     VMS_std_crypt(void);

/*
 * Returns the salt for VMS_std_set_salt().
 */
	extern char    *VMS_std_get_salt(char *ciphertext);

/*
 * Converts an ASCII ciphertext to binary.
 */
	extern VMS_word *VMS_std_get_binary(char *ciphertext);

#endif
#ifndef VMS
#ifndef HAVE_PTHREADS
#define NOPTHREADS
#endif
#endif


#define FORMAT_LABEL			"openvms"
#define FORMAT_NAME			"OpenVMS Purdy"

#ifdef NOPTHREADS
#define BENCHMARK_COMMENT		" uniprocessor"
#else
#define BENCHMARK_COMMENT		" multithreaded"
#endif
#define BENCHMARK_LENGTH		-1


#define PLAINTEXT_LENGTH		31
#define CIPHERTEXT_LENGTH		UAF_ENCODE_SIZE

#define BINARY_SIZE			8
#define SALT_SIZE			12

#define MIN_KEYS_PER_CRYPT		(SAVE_LIMIT/2)
#define MAX_KEYS_PER_CRYPT		SAVE_LIMIT

	static struct fmt_tests tests[] = {
		{"$V$S44zI913bBx-UJrcFSC------D", "President#44"},
		{"$V$9AYXUd5LfDy-aj48Vj54P-----", "USER"},
		{"$V$p1UQjRZKulr-Z25g5lJ-------", "service"},
		{NULL}
	};

	static int      saved_keypos[MAX_KEYS_PER_CRYPT];
	static char     saved_key[MAX_KEYS_PER_CRYPT][PLAINTEXT_LENGTH + 1];
	struct result_hash *VMS_std_crypt_out;
	int             VMS_dbg_flag;
#define DBGPR if ( VMS_dbg_flag == 0 ); else printf
/*
 * See if signature of ciphertext (from passwd file) matches the hack
 * produced by the uaf_encode routine (starts with $V$)
 */
	static int      valid(char *ciphertext, struct fmt_main * self)
{
	if (strncmp(ciphertext, "$V$", 3))
		return 0;	/* no match */

	if (strlen(ciphertext) < 18)
		return 0;

	return 1;		/* Only one split */
}
/*
 * Prepare function.
 */
char           *
prepare(char *split_fields[10], struct fmt_main * pFmt)
{
	return split_fields[1];
}
/*
 * Hash the internal representation of a ciphertext hash to 4, 8, or 12 bits.
 */
static int 
binary_hash_0(void *binary)
{
	unsigned short int *bw;
	bw = binary;
	DBGPR("/VMS_FMT/ binary_hash_0(%x-%x-%x-%x)\n", bw[0], bw[1], bw[2], bw[3]);
	return ((bw[0] ^ bw[1]) ^ (bw[2] ^ bw[3])) & 0x0F;
}

static int 
binary_hash_1(void *binary)
{
	unsigned short int *bw;
	bw = binary;
	DBGPR("/VMS_FMT/ binary_hash_1(%x-%x-%x-%x)\n", bw[0], bw[1], bw[2], bw[3]);
	return ((bw[0] ^ bw[1]) ^ (bw[2] ^ bw[3])) & 0x0FF;
}

static int 
binary_hash_2(void *binary)
{
	unsigned short int *bw;
	bw = binary;
	DBGPR("/VMS_FMT/ binary_hash_2(%x-%x-%x-%x)\n", bw[0], bw[1], bw[2], bw[3]);
	return ((bw[0] ^ bw[1]) ^ (bw[2] ^ bw[3])) & 0x0FFF;
}
static int 
binary_hash_3(void *binary)
{
	unsigned short int *bw;
	bw = binary;
	DBGPR("/VMS_FMT/ binary_hash_3(%x-%x-%x-%x)\n", bw[0], bw[1], bw[2], bw[3]);
	return ((bw[0] ^ bw[1]) ^ (bw[2] ^ bw[3])) & 0x0FFFF;
}
static int 
binary_hash_4(void *binary)
{
	unsigned short int *bw;
	bw = binary;
	DBGPR("/VMS_FMT/ binary_hash_4(%x-%x-%x-%x)\n", bw[0], bw[1], bw[2], bw[3]);
	return ((bw[0] ^ bw[1]) ^ (bw[2] ^ bw[3])) & 0x0FFFFF;
}

/*
 * Hash the internal representation of save crypt_all result binary
 * to 4, 8, 12, 16, or 20 bits.

#define PASSWORD_HASH_SIZE_0            0x10
#define PASSWORD_HASH_SIZE_1            0x100
#define PASSWORD_HASH_SIZE_2            0x1000
#define PASSWORD_HASH_SIZE_3            0x10000
#define PASSWORD_HSIZE_4            0x100000
 */
static int 
get_hash_0(int index)
{
	if (!VMS_std_crypt_out)
		return 0;
	DBGPR("/VMS_FMT/ get_hash_0(%d)\n", index);
	return binary_hash_0(&VMS_std_crypt_out[index].info.hash);
}

static int 
get_hash_1(int index)
{
	if (!VMS_std_crypt_out)
		return 0;
	DBGPR("/VMS_FMT/ get_hash_1(%d)\n", index);
	return binary_hash_1(&VMS_std_crypt_out[index].info.hash);
}

static int 
get_hash_2(int index)
{
	if (!VMS_std_crypt_out)
		return 0;
	DBGPR("/VMS_FMT/ get_hash_2(%d)\n", index);
	return binary_hash_2(&VMS_std_crypt_out[index].info.hash);
}

static int 
get_hash_3(int index)
{
	if (!VMS_std_crypt_out)
		return 0;
	DBGPR("/VMS_FMT/ get_hash_3(%d)\n", index);
	return binary_hash_3(&VMS_std_crypt_out[index].info.hash);
}

static int 
get_hash_4(int index)
{
	if (!VMS_std_crypt_out)
		return 0;
	DBGPR("/VMS_FMT/ get_hash_4(%d)\n", index);
	return binary_hash_4(&VMS_std_crypt_out[index].info.hash);
}

/*
 * Hash the internal representation of salt (12 bytes) to 10 bits.
 */
static int 
salt_hash(void *salt)
{
	unsigned short int *bw;
	bw = salt;
	return ((bw[0] ^ bw[1]) ^ (bw[2] ^ bw[3]) ^ bw[4]) & 0x3FF;
}
/*
 * Save a password (key) for testing.  VMS_std_set_key returns position value
 * we can use if needed to recall the key by a fmt->get_key request.  On get_key
 * return a private copy.
 */
static void 
set_key(char *key, int index)
{
	saved_keypos[index] = VMS_std_set_key(key, index);
	if ((*key == 0) && (saved_keypos[index] == 0)) {
		/* Special sequence, 'delete' existing key' */
		VMS_std_crypt();
	}
}

static char    *
get_key(int index)
{
	char           *key = VMS_std_get_key(saved_keypos[index], index);

	return strcpy(saved_key[index], key);
}

static int 
cmp_one(void *binary, int index)
{
	uaf_qword      *qw;
	if (!VMS_std_crypt_out)
		return 0;

	qw = binary;

	DBGPR("/VMT_FMT/ cmp_one comparing %llx against saved (%llx) [%d]\n",
		*qw, VMS_std_crypt_out[index].info.hash, index);
	return UAF_QW_EQL(*qw, VMS_std_crypt_out[index].info.hash);
}

static int 
cmp_all(void *binary, int count)
{
	int             i;
	uaf_qword      *qw;
	uaf_lword      *hint, ndx;
	if (!VMS_std_crypt_out)
		return 0;

	qw = binary;
	/*
         * Check the summary hint.  A possible hit would set the hash table entry
         * to the current generation number so we should probably look for match.
         */
	hint = &VMS_std_hint->seen[UAF_QW_AND(*qw, VMS_std_hint->mask)];
	i = (VMS_std_hint->seq == (*hint & HINT_GENERATION_MASK));
	DBGPR("/VMT_FMT/ cmp_all comparing %llx in saved [0..%d] (%llx): %d\n",
		*qw, count - 1, *qw & VMS_std_hint->mask, i);
	if (!i)
		return 0;
	/*
         * Possible hit, extract index and compare if we can.
         */
	ndx = (*hint) >> HINT_GENMASK_BITS;
	if (ndx > SAVE_LIMIT) {
		return 1;	/* multiple hits, can't single out one to
				 * test. */
	}
	return cmp_one(binary, ndx);
}

static int 
cmp_exact(char *source, int index)
{
	uaf_qword      *qw;

	if (!VMS_std_crypt_out)
		return 0;
	qw = (uaf_qword *) VMS_std_get_binary(source);
	/*
         * Note that comparison depends upon hash being first member of
         * uaf_hash_info structure.
         */
	return UAF_QW_EQL(*qw, VMS_std_crypt_out[index].info.hash);
}
static char    *
fmt_vms_split(char *ciphertext, int index)
{
	return ciphertext;
}
static void 
fmt_vms_clear_keys(void)
{
}
/*
 * Class record.
 */
struct fmt_main fmt_openvms = {
	{
		FORMAT_LABEL,
		FORMAT_NAME,
		VMS_ALGORITHM_NAME,
		BENCHMARK_COMMENT,
		BENCHMARK_LENGTH,
		PLAINTEXT_LENGTH,
		BINARY_SIZE,
		SALT_SIZE,
		MIN_KEYS_PER_CRYPT,
		MAX_KEYS_PER_CRYPT,
		FMT_CASE | FMT_8_BIT | FMT_BS,
		tests
	}, {
		VMS_std_init,
		prepare,
		valid,
		fmt_vms_split,
		(void *(*) (char *)) VMS_std_get_binary,
		(void *(*) (char *)) VMS_std_get_salt,
		{
			binary_hash_0,
			binary_hash_1,
			binary_hash_2,
			binary_hash_3,
			binary_hash_4
		},
		salt_hash,
		(void (*) (void *)) VMS_std_set_salt,
		set_key,
		get_key,
		fmt_vms_clear_keys,
		(void (*) (int)) VMS_std_crypt,
		{
			get_hash_0,
			get_hash_1,
			get_hash_2,
			get_hash_3,
			get_hash_4
		},
		cmp_all,
		cmp_one,
		cmp_exact
	}
};
/*
 * Special function that DLL will exported, called by --dynamic-fmt option.
 */
struct fmt_main *
FMT_LOADER(int version)
{
	/*
         * Only allow the load if caller's version matches the one we were
         * compiled with.
         */
	if (version == FMT_MAIN_VERSION)
		return &fmt_openvms;

	fprintf(stderr, "Failed to activate %s format due to version mismatch.\n",
		fmt_openvms.params.label);
	fprintf(stderr, "   requested version: %d, supported by DLL: %d\n",
		version, FMT_MAIN_VERSION);

	return 0;
}
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*
 * This file is part of John the Ripper password cracker,
 *
 * Using md5_std.c as a starting point, make version for OpenVMS password
 * hash.  This version uses Posix threads to perform the hashing in
 * parallel. Compile with macro NOPTHREADS defined to inhibit multi-threading.
 *
 * Environment variables examined at initialization:
 *    JOHN_OPENVMS_THREADS  Dispatch parameters for threading, value is
 *                          of for "n m l" where:
 *                            n - Number of threads to create (default=1).
 *                            m - Min. batch size for nonblocked start (def=32).
 *                            l - Max. batch size for worker thread (def=60).
 *    JOHN_OPENVMS_DEBUG n  Output an exectution trace if non-zero.
 *
 * Limit and threshold value can have a maximum of the module SAVE_LIMIT
 * setting, which is 120.
 *
 * Author:  David Jones
 * Date:    9-JUL-2009
 * Revised: 17-JUL-2009     Cleanup, make work under OS X.
 * Revised: 26-JUL-2009     Cleanup comments.
 * Revised: 31-JUL-2009     Fix uaf_qword reference for NOLONGLONG compiles.
 * Revised:  2-AUG-2009     Eliminate redundant update of hint table entries.
 * Revised: 27-AUG-2011     Remove log_event calls.
 *
 */
#include <string.h>
#ifdef NOPTHREADS
#include <time.h>
#define MUTEX long
#define LOCK_MUTEX(lck)
#define UNLOCK_MUTEX(lck)
#define INIT_MUTEX(lck)
#define CONDVAR  long
#define THREAD_ATTR long
#else
#include <pthread.h>
#define MUTEX pthread_mutex_t
#define LOCK_MUTEX(lck) pthread_mutex_lock(lck)
#define UNLOCK_MUTEX(lck) pthread_mutex_unlock(lck)
#define INIT_MUTEX(lck) pthread_mutex_init(lck, 0)
#define CONDVAR pthread_cond_t
#define THREAD_ATTR pthread_attr_t
#endif

#ifdef VMS
#include <uaidef.h>
#include <descrip.h>
#include <lib$routines.h>
#include <times.h>
#else
#include <sys/times.h>
#endif
#ifndef UAI$M_PWDMIX
#define UAI$M_PWDMIX 0x2000000
#endif

#ifndef TRY			/* fallback if excpetions not support */
#define TRY if ( wq )
#define CATCH_ALL else
#define THIS_CATCH 0
#define RERAISE printf("No exception support")
#define ENDTRY
#define pthread_exc_report_np(x) printf("Exception %d caught\n",(int)(x))
#define pthread_exc_get_status_np(c,v) (*(v)=0)
#elif !defined(VMS)
#define pthread_exc_report_np(x) printf("Exepction %d caught\n", (int)(x))
#define pthread_exc_get_status_np(c,v) (*(v)=0)
#endif

/*
 * Static storage for data preserved across calls. Crypt_out and summary
 * are shared with VMS_FMT module (by way of global pointers).
 * Summary structure is used by cmp_all to quickly reject non-matches.
 */
static struct result_hash crypt_out[SAVE_LIMIT];
static struct result_hint *summary;
struct result_hint *VMS_std_hint = 0;
struct result_hash *VMS_std_crypt_out = 0;

int             VMS_dbg_flag;
static void     report_statistics(void);	/* forward reference */
#define STATS_INTERVAL 10000	/* Crypt calls between stats report */
#define DBGPR if ( VMS_dbg_flag == 0 ); else printf
#define DEFAULT_WORKER_COUNT "1"
#define DEFAULT_PENDING_THRESHOLD "66"
#define DEFAULT_PENDING_LIMIT "120"
#define PLAINTEXT_LENGTH 31
#define WO_STS_DEFUNCT 0
#define WO_STS_BUSY 1
#define WO_STS_COMPLETED 2

/*
 * Work_item struct saves arguments from a set_key request.
 */
struct work_item {
	int             ndx;	/* Index of output location in crypt_out */
	char            key[PLAINTEXT_LENGTH + 1];	/* candidate password */
};
/*
 * The work_queue structure collects set_key/set_salt requests and tracks
 * the assignment of blocks of pending work items to worker threads.  The
 * work_order structure coordinates this handoff been the main thread and
 * a single worker thread.
 */
struct work_queue {
	MUTEX           lock;	/* Protects structures for shared access */
	CONDVAR         order_done;	/* Signaled by worker at  BUSY ->
					 * COMPLETED */
	int             worker_count;	/* number of threads (size of
					 * work_order[]) */
	int             workers_active;	/* Number of worker threads with BUSY
					 * status */

	struct {		/* Statistics for VMS_std_crypt() calls */
		long            calls;	/* Number of times crypt function
					 * called */
		long            mile_marker;	/* Report stats when calls ==
						 * mile_marker */
		long            keys_computed;
		long            preempts;	/* number of times
						 * dispatch_pending preempts */
		long            preempt_idle;	/* Counts no active workers
						 * after preempt */
		clock_t         start_clock, cur_clock;
		struct tms      start_cpu, cur_cpu;
	}               crypt_counters;
	struct uaf_hash_info salt;	/* Salt for new orders */
	int             salt_valid;	/* true if salt set since last crypt */
	int             keys_changed;	/* true if any keys set since last
					 * crypt */
	int             request_count;	/* Requests pending at current
					 * sequence */
	int             dispatched;	/* Number of pending assigned to
					 * workers */
	int             pending_threshold;	/* Count at which to create
						 * work order */
	int             pending_limit;	/* Max. count allowed in work order */

	struct work_order *order;	/* Array of wq->worker_count size */

	struct work_item request[SAVE_LIMIT];	/* Candidate passwords to
						 * encrypt */
};

struct work_order {
	struct work_queue *wq;
	int             assignee;	/* thread number assigned to do work */
	int             status;	/* 0-initial, 1-busy, 2-completed */
	CONDVAR         order_ready;	/* Signaled when status changed to
					 * BUSY */

	struct uaf_hash_info salt;	/* Salt to use to process items */

	struct work_item *first;/* Points into wq->request array. */
	int             count;	/* Number of items to process */
	char            scratch[8140];
};
static struct work_queue *wq;
/*
 * Process request item[0..count-1] array, placing result of hashes in the
 * crypt_out array.
 * Caller must hold held_lock mutex, which we release and re-acquire.
 */
static void 
compute_hashes(int assignee, MUTEX * held_lock,
	struct uaf_hash_info * salt, struct work_item * item, int count)
{
	int             ndx, match;
	uaf_lword       smask, info;
	struct work_item *cur_item, *end_item;
	struct {		/* Array to save results for */
		int             ndx;	/* inclusion in summary table */
		uaf_lword      *seen_entry;	/* points into seen array */
	}              *resptr, result[count + 1];
	/*
         * Compute hash of each item[] password using the same input salt.  The
         * test_password  function saves resulting hash back to the uaf_hash_info
         * structure (1st argument) when the replace_if argument is 1 (true).
         */
	UNLOCK_MUTEX(held_lock);
	end_item = &item[count];/* one past final item to process */
	resptr = result;	/* Saves result for insert in summary */
	smask = summary->mask;
	for (cur_item = item; cur_item < end_item; cur_item++) {
		ndx = cur_item->ndx;	/* Output location in crypt_out[] */
		crypt_out[ndx].info = *salt;	/* set salt for hash */
		match = uaf_test_password(&crypt_out[ndx].info, cur_item->key, 1);
		resptr->ndx = ndx;
		resptr->seen_entry =
			&summary->seen[UAF_QW_AND(crypt_out[ndx].info.hash, smask)];
		resptr++;
	}
	/*
         * Update summary table using nxtseq value as generation number.
         * Table entries are relatively close, so keep summary->lock mutex entire
         * time.
         */
	DBGPR("/VMS_STD-%d/ updating %d hint table entries.\n", assignee, count);
	LOCK_MUTEX(summary->lock_vp);
	if (summary->nxtseq == 0) {
		/*
		 * We wrapped, reset table and start seq at 1 (zero never
		 * used)
		 */
		memset(summary->seen, 0, (summary->mask + 1) * sizeof(uaf_lword));
		summary->nxtseq = 1;
	}
	while ((--resptr) >= result) {
		info = (*resptr->seen_entry) & HINT_GENERATION_MASK;
		ndx = resptr->ndx;
		if (info == summary->nxtseq)
			ndx = HINT_MULTIPLE;	/* double hit */
		*resptr->seen_entry = summary->nxtseq | (ndx << HINT_GENMASK_BITS);
	}
	UNLOCK_MUTEX(summary->lock_vp);
	LOCK_MUTEX(held_lock);
}
#ifndef NOPTHREADS
/*
 * Thread start routine for a worker thread.
 */
static void    *
worker_start(void *work_order_vp)
{
	struct work_order *wo;	/* Work_order_vp as non void * */
	/*
         * Main loop.  Assume initial status of BUSY with zero work items, main
         * thread can't add items to the work order until we mark it completed.
         */
	wo = work_order_vp;
	LOCK_MUTEX(&wo->wq->lock);
	TRY
	do {
		/*
		 * Hash the designated subset of wq->request (count items starting
		 * at wo->first).  Lock will be released and re-acquired.
		 */
		compute_hashes(wo->assignee, &wo->wq->lock,
			&wo->salt, wo->first, wo->count);
		/*
		 * Notify main thread via order_done condition that we finished.
		 */
		wo->wq->workers_active--;
		wo->status = WO_STS_COMPLETED;
		if (0 != pthread_cond_signal(&wo->wq->order_done)) {
			fprintf(stderr, "Worker start internal error, cond_signal\n");
			return wo;
		}
		/*
		 * Wait for main thread to change work order's state back to BUSY,
	         * indicating it has updated it with new input arguments.
		 * Thread will be shut down if marked busy and assignee number cleared.
		 */
		do
			if              (0 != pthread_cond_wait(&wo->order_ready, &wo->wq->lock)) {
				fprintf(stderr, "Condition wait error, abort\n");
				UNLOCK_MUTEX(&wo->wq->lock);
				return wo;
		} while (wo->status != WO_STS_BUSY);

	} while (wo->assignee);
	CATCH_ALL {
		unsigned int    stsval;
		printf("Exception in worker thread %d\n", wo->assignee);
		if (pthread_exc_get_status_np(THIS_CATCH, &stsval) == 0) {
			printf("Status code: %d\n", stsval);
		} else
			printf("Status retrieval failed\n");
		pthread_exc_report_np(THIS_CATCH);
		RERAISE;
	}
	ENDTRY
		UNLOCK_MUTEX(&wo->wq->lock);
	return wo;
}
#endif
/*
 * Process work queue request items not yet dispatched to workers.
 */
static void 
dispatch_pending(const char *caller, int wait_for_completion)
{
	int             backlog, lot_size;
	struct work_order *wo;
	/*
         * Assign blocks of pending request items to worker threads until
         * no more backlog, no workers, or on final chunk and wait requested.
         */
	LOCK_MUTEX(&wq->lock);
#ifndef NOPTHREADS
	for (backlog = wq->request_count - wq->dispatched;
		(backlog > 0) && (wq->worker_count > 0);
		backlog = wq->request_count - wq->dispatched) {
		/*
		 * Limit number of items (lot_size) to handoff to worker.
		 */
		if (backlog > wq->pending_limit)
			lot_size = wq->pending_limit;
		else if (wait_for_completion)
			break;	/* do final lot in this thread */
		else
			lot_size = backlog;	/* assigne final lot to worker */
		/*
		 * Wait for a free worker, then find first available work order.
		 */
		while (wq->workers_active >= wq->worker_count) {
			if (0 != pthread_cond_wait(&wq->order_done, &wq->lock)) {
				fprintf(stderr, "Condition wait error, abort\n");
				return;
			}
		}
		for (wo = wq->order; wo->status != WO_STS_COMPLETED; wo++) {
			if (wo == &wq->order[wq->worker_count - 1]) {
				/*
				 * Shouldn't happen since workers_active <
				 * worker_count
				 */
				fprintf(stderr, "Bugcheck, queue state inconsistency\n");
				exit(44);
			}
		}
		/*
		 * Fill in work order details and hand off to worker thread.
		 */
		wo->status = WO_STS_BUSY;
		wo->first = &wq->request[wq->dispatched];	/* First undispatched */
		wo->count = lot_size;
		wo->salt = wq->salt;

		pthread_cond_signal(&wo->order_ready);
		wq->dispatched += lot_size;
		wq->workers_active++;
		DBGPR("/VMS_STD/ initiated %s work order #%d, count: %d (%s)\n",
			wait_for_completion ? "synch" : "async", wo->assignee, wo->count, caller);
	}
	/*
         * Compute any remaining undispatched requests directly from this thread.
         * Compute_hashes releases and re-acquires lock so signalled workers can
         * proceed while we compute.  Track the number of times we did this.
         */
	if (backlog > 0) {
		DBGPR("/VMS_STD/ preempted final work order, count: %d %s\n",
			backlog, caller);
		compute_hashes(0, &wq->lock, &wq->salt,
			&wq->request[wq->dispatched], backlog);
		wq->dispatched += backlog;
		wq->crypt_counters.preempts++;
		if (wq->workers_active == 0)
			wq->crypt_counters.preempt_idle++;
	}
	/*
         * Wait for all work orders to be completed if caller requested a stall.
         */
	if (wait_for_completion)
		while (wq->workers_active > 0) {
			if (0 != pthread_cond_wait(&wq->order_done, &wq->lock)) {
				fprintf(stderr, "Condition wait error, abort\n");
				return;
			}
		}
#else
	backlog = wq->request_count - wq->dispatched;
	lot_size = backlog;
	wo = wq->order;
	DBGPR("/VMS_STD/ non-threaded work queue dispatch(%p), count: %d %s\n",
		wo, backlog, caller);
	compute_hashes(0, &wq->lock, &wq->salt,
		&wq->request[wq->dispatched], backlog);
	wq->dispatched += lot_size;
	wq->crypt_counters.preempts++;
	if (wq->workers_active == 0)
		wq->crypt_counters.preempt_idle++;

#endif
	UNLOCK_MUTEX(&wq->lock);

}
/*
 * Initialization function for module, called by main program.
 */
void 
VMS_std_init(struct fmt_main * self)
{
	char           *env, *arg;
	int             workers, threshold, limit, smask;
#ifndef NOPTHREADS
	int             i;
	static THREAD_ATTR attr;
#endif
	MUTEX          *lock_p;
	/*
         * Read configuration from environment variables.
         */
	env = getenv("JOHN_OPENVMS_DEBUG");	/* set debug mode */
	if (env)
		VMS_dbg_flag = atoi(env);

	env = getenv("JOHN_OPENVMS_THREADS");
	arg = env ? strtok(env, " ") : (char *) 0;
	workers = atoi(arg ? arg : DEFAULT_WORKER_COUNT);
	if (arg)
		arg = strtok(0, " ");
	threshold = atoi(arg ? arg : DEFAULT_PENDING_THRESHOLD);
	if (arg)
		arg = strtok(0, " ");
	limit = atoi(arg ? arg : DEFAULT_PENDING_LIMIT);
	/*
         * Initialize the structure for storing result hints (speeds up cmp_all).
         * Structure ends in seen array which is variably sized to be at least
         * 200 times larger than SAVE_LIMIT.  cmp_all should get a false positive
         * no more than 1% of the time.
         */
	lock_p = malloc(sizeof(MUTEX));
	for (smask = 255; smask < (SAVE_LIMIT * 200);)
		smask = (smask << 1) | 1;
	summary = calloc(1, sizeof(struct result_hint) + (smask * sizeof(uaf_lword)));
	summary->mask = smask;
	VMS_std_hint = summary;
	/*
         * Initialize for multi-threading.  Create crypting threads, one for
         * each processor and assign a thread number 1..N.
         */
	uaf_init();
	summary->lock_vp = lock_p;	/* hint struct uses opaque pointer */
	INIT_MUTEX(summary->lock_vp);

	wq = malloc(sizeof(struct work_queue));
	wq->worker_count = (workers > 0) ? workers : 0;
	wq->order = malloc(sizeof(struct work_order) * (wq->worker_count + 1));
	wq->pending_limit = (limit < SAVE_LIMIT) ? limit : SAVE_LIMIT;
	wq->pending_threshold = (threshold < limit) ? threshold : limit;
	INIT_MUTEX(&wq->lock);
#ifndef NOPTHREADS
	pthread_cond_init(&wq->order_done, 0);
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

	for (i = 0; i < wq->worker_count; i++) {
		pthread_t       tid;
		memset(&wq->order[i], 0, sizeof(struct work_order));
		wq->order[i].wq = wq;	/* back link */
		wq->order[i].status = WO_STS_BUSY;	/* in_progress, worker
							 * completes */
		wq->order[i].assignee = i + 1;
		wq->workers_active++;
		pthread_cond_init(&wq->order[i].order_ready, 0);

		if (0 != pthread_create(&tid, &attr, worker_start, &wq->order[i])) {
			fprintf(stderr, "Error creating worker thread #%d\n", i + 1);
		}
	}
#else
	wq->worker_count = 0;	/* Force skipping thread scheduling */
#endif
	report_statistics();	/* Initialize crypt_counters */
	memset(&wq->crypt_counters, 0, sizeof(wq->crypt_counters));
}
/*
 * Save key (password) for producing ciphertext (hash) from it and saved salt.
 */
int 
VMS_std_set_key(char *key, int index)
{
	int             unbusy, request_count;
	/*
         * Validate index argument in range and reset request_count if first
         * set_key after a crypt call.  This lets caller change the salt and
         * call crypt again to compute new hashes with the same set of saved keys.
         */
	if ((index < 0) || (index >= SAVE_LIMIT)) {
		fprintf(stderr, "BUGCHECK, vms_std_set_key: index=%d out of range", index);
		return 0;
	}
	if (!wq->keys_changed) {
		wq->keys_changed = 1;
		wq->request_count = 0;
		wq->dispatched = 0;
	}
	request_count = wq->request_count;
	if (request_count >= SAVE_LIMIT) {
		printf("Bugcheck, too many set_keys, req=%d, limit=%d!\n",
			request_count, SAVE_LIMIT);
		request_count = SAVE_LIMIT - 1;
	}
	/*
         * Add key and index (output location) to list of pending items.
         */
	wq->request[request_count].ndx = index;
	if (strlen(key) > PLAINTEXT_LENGTH) {
		memcpy(wq->request[request_count].key, key, PLAINTEXT_LENGTH);
		wq->request[request_count].key[PLAINTEXT_LENGTH] = 0;
	} else
		strcpy(wq->request[request_count].key, key);
	DBGPR("/VMS_STD/ saved key[%d] as item[%d] '%s'\n", index,
		request_count, wq->request[request_count].key);
	wq->request_count = request_count + 1;
	/*
         * If a salt has been set since last crypt operation, see if we should
         * start clearing the list with another thread.
         */
	if (wq->salt_valid) {
		unbusy = wq->request_count - wq->dispatched;
		if (unbusy >= wq->pending_threshold)
			dispatch_pending("set key", 0);
	}
	return request_count;	/* Position key saved at in wq->request array */
}
/*
 * Save salt for producing ciphertext from it and saved keys at next crypt call.
 */
void 
VMS_std_set_salt(void *salt_vp)
{
	int             pending;
	/*
         * If salt changes twice between std_crypt calls, complete hashing of
         * pending keys with the old salt.
         */
	if (wq->salt_valid && wq->keys_changed)
		dispatch_pending("set salt 2", 1);
	wq->salt_valid = 1;
	/*
         * Save salt in work queue structure and decode back username and flags.
         */
	memcpy(&wq->salt.salt, salt_vp, 12);
	uaf_packed_convert(&wq->salt.username, 0);
	wq->salt.flags = (wq->salt.opt & 1) ? UAI$M_PWDMIX : 0;
	VMS_std_crypt_out = crypt_out;	/* make availabe to VMS_FMT module */
	/*
         * Determine if pending set_keys have accumulated enough to cut a work
         * order for worker threads to begin hash operations in parallel.
         */
	pending = (wq->request_count - wq->dispatched);
	if (!wq->keys_changed)
		pending = (-pending);
	DBGPR("/VMS_STD/ set salt '%s'(alg=%d), pending keys: %d\n",
		wq->salt.username.s, wq->salt.alg, pending);

	if (pending >= wq->pending_threshold)
		dispatch_pending("set salt", 0);

}
/*
 * Hash the password and salt saved with VMS_std_set_key and VMS_std_set_salt,
 * saving the result in global storage for retrieval by vms_fmt.c module.
 */
void 
VMS_std_crypt(void)
{
	DBGPR("/VMS_STD/ crypting to complete open work (%d of %d dispatched)\n",
		wq->dispatched, wq->request_count);
	/*
         * Finish any set_key requests pending.  Note that if set_salt changed the
         * salt with no key changes, the entire block of keys will be re-computed.
         */
	dispatch_pending("crypt", 1);
	/*
         * Bump the generation number for the hint.  We use generation number to
         * mark hint table entries to avoid clearing the entire array each time.
         */
	LOCK_MUTEX(summary->lock_vp);
	summary->seq = summary->nxtseq;
	summary->nxtseq = (summary->nxtseq + 1) & HINT_GENERATION_MASK;
	UNLOCK_MUTEX(summary->lock_vp);
	/*
         * Keep statistics on crypt() activity.
         */
	wq->crypt_counters.calls++;
	wq->crypt_counters.keys_computed += wq->request_count;
	if (wq->crypt_counters.calls >= wq->crypt_counters.mile_marker) {
		report_statistics();
	}
	/*
         * Reset salt_valid and keys_changed so we don't proactively recalculate
         * until both have been received (or crypt called again after one received).
         */
	wq->salt_valid = 0;
	wq->dispatched = 0;	/* All requests need re-processed */
	wq->keys_changed = 0;	/* defer reseting request_count until change */

	return;
}
/*
 * Return pointer to saved key for position.  Do sanity check with index.
 */
char           *
VMS_std_get_key(int position, int index)
{
	if (wq->request[position].ndx == index)
		return wq->request[position].key;

	printf("BUGCHECK, wrong index at saved_key position\n");
	return "";
}
/*
 * Extract salt from ciphertext string to static storage and return
 * pointer to it.  Salt is effectively 70-80 bits (username, salt,
 * algorithm, pwdmix flag).
 */
char           *
VMS_std_get_salt(char *ciphertext)
{
	static struct uaf_hash_info pwd;

	uaf_hash_decode(ciphertext, &pwd);

	DBGPR("/VMS_STD/ get_salt decoded '%s' to %x/%x-%x-%x-%x-%x\n",
		ciphertext, pwd.salt, pwd.alg, pwd.username.r40[0], pwd.username.r40[1],
		pwd.username.r40[2], pwd.username.r40[3]);
	return (char *) &pwd.salt;
}
/*
 * Extract binary hash from ciphertext into static storage and return
 * pointer to it.
 */
VMS_word       *
VMS_std_get_binary(char *ciphertext)
{
	static union {
		struct uaf_hash_info pwd;
		VMS_word        b[16];
	}               out;

	uaf_hash_decode(ciphertext, &out.pwd);
	DBGPR("/VMS_STD/ get_binary decoded '%s' to %lx->%x.%x\n",
		ciphertext, (unsigned long int) out.b, out.b[1], out.b[0]);

	return out.b;
}
/*
 * Calculate program performance and save in job logical name table.
 */
static void 
report_statistics(void)
{
	long long       elapsed;
	long            cpu_ticks;
	static char     logical[120], line[120];
#ifdef VMS
	int             code, status;
	static          $DESCRIPTOR(logical_dx, logical);
	static          $DESCRIPTOR(lognam_dx, "JOHN_PERFORMANCE_STATS");
	static          $DESCRIPTOR(table_dx, "LNM$JOB");
#endif

	if (wq->crypt_counters.mile_marker == 0) {	/* Initial call */
		wq->crypt_counters.start_clock = times(&wq->crypt_counters.start_cpu);
		/*
		 * log_event ( "OpenVMS crypt() performace counters
		 * initialized." );
		 */
		sprintf(line, "Threads: %d %d %d",
			wq->worker_count, wq->pending_threshold, wq->pending_limit);
		/* log_event ( "OpenVMS crypt() %s", line ); */
	} else {
		/*
		 * Use system timers to get clock and CPU time, then make logical
		 * name with total stats.
		 */
		struct tms      prev_cpu;

		prev_cpu = wq->crypt_counters.cur_cpu;
		wq->crypt_counters.cur_clock = times(&wq->crypt_counters.cur_cpu) -
			wq->crypt_counters.start_clock;
		wq->crypt_counters.cur_cpu.tms_utime -=
			wq->crypt_counters.start_cpu.tms_utime;
		wq->crypt_counters.cur_cpu.tms_stime -=
			wq->crypt_counters.start_cpu.tms_stime;

		elapsed = wq->crypt_counters.cur_clock;
		cpu_ticks = wq->crypt_counters.cur_cpu.tms_utime;

		sprintf(logical,
			"Elapsed: %lld CPU: %ld crypts: %ld keys: %ld preempts: %ld(%ld) thresh %d",
			elapsed, cpu_ticks, wq->crypt_counters.calls,
			wq->crypt_counters.keys_computed, wq->crypt_counters.preempts,
			wq->crypt_counters.preempt_idle, wq->pending_threshold);

		/* log_event ( "OpenVMS crypt() %s", logical ); */
#ifdef VMS
		logical_dx.dsc$w_length = strlen(logical);
		status = LIB$SET_LOGICAL(&lognam_dx, &logical_dx, &table_dx, 0, 0);
		if ((status & 1) == 0) {
			fprintf(stderr, "Set logical error: %d", status);
			fprintf(stderr, "Perf_report: %d '%s'\n", status, logical);
		}
#endif
		wq->crypt_counters.preempts = 0;
		wq->crypt_counters.preempt_idle = 0;
	}
	wq->crypt_counters.mile_marker += STATS_INTERVAL * (VMS_dbg_flag) ? 1 : 20;
}
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*
 * Define functions to encode/deocde SYSUAF password information as a
 * printable string suitable for use in a UNIX-style passwd file (i.e. no
 * colons).
 *
 * The password string is 18 characters of the form:
 *
 *    $V$ssshhhhhhhhhhhh
 *
 * where:
 *    $V$	    Tags field as VMS SYSUAF format.
 *    sss           Encodes the 16 bit salt plus 2 flag bits:
 *                     <0>  If set, mixed case password in use.
 *                     <1>  Reserved.
 *    hhhhhhhhhhhh  Encodes 8-bit algorithm code plus 64-bit hash value.
 *
 * Encoding is 6 bits per character:
 *
 *    0000000000111111111122222222223333333333444444444455555555556666
 *    0123456789012345678901234567890123456789012345678901234567890123
 *    -ABCDEFGHIJKLMNOPQRTSUVWXYZ0123456789abcedfghijklmnopqrstuvwxyz+
 *
 * Author:	David Jones
 * Date:	11-JUL-2209
 * Revised:	14-AUG-2009	Initialize P in main thread.
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#ifdef VMS
#include <descrip.h>
#include <uaidef.h>
#include <starlet.h>
#else
/*
 * Emulate symbols defined for VMS services.
 */
#define UAI$C_AD_II 0
#define UAI$C_PURDY 1
#define UAI$C_PURDY_V 2
#define UAI$C_PURDY_S 3

struct dsc$descriptor_s {
	unsigned short int dsc$w_length;
	unsigned char   dsc$b_dtype, dsc$b_char;
	char           *dsc$a_pointer;
};
#define $DESCRIPTOR(x,s) struct dsc$descriptor_s x={sizeof(s), 1, 1, s}
#endif

#ifndef UAI$M_PWDMIX
#define UAI$M_PWDMIX 0x2000000
#endif
#ifdef NOPTHREADS
typedef int     pthread_mutex_t;
#define PTHREAD_MUTEX_INITIALIZER 0
#define pthread_mutex_lock ( pthread_mutex_t * )
#define pthread_mutex_unlock ( pthread_mutex_t * )
#else
#include <pthread.h>
#endif				/* NOPTHREADS */

/*
 * Declare static globals that don't change once initialized as well as
 * pthread objects used to sychronize access and/or initialization.
 */
static pthread_mutex_t uaf_static = PTHREAD_MUTEX_INITIALIZER;
static const char *enc_set =
"-ABCDEFGHIJKLMNOPQRTSUVWXYZ0123456789abcedfghijklmnopqrstuvwxyz+";
static const char *r50_set = " ABCDEFGHIJKLMNOPQRSTUVWXYZ$._0123456789";
#define SYSUAF_FMT "$V$%64q%8b%16w%64T%4b"
static int      enc_map_ready = 0;
static unsigned short enc_map[256];
static unsigned short r50_map[256];
/*
 * Forward reference to main function that hashes a password as OpenVMS does
 * in its SYSUAF file.  Function takes same arguments as SYS$HASH_PASSWORD
 * system service but in different order.
 */
static int 
hash_password(uaf_qword *,	/* Receives result hash */
	struct dsc$descriptor_s *,	/* Password */
	unsigned char, unsigned short,	/* Algorithm code and salt */
	struct dsc$descriptor_s *);	/* Username (eff. more salt) */

/****************************************************************************/
/*
 * Internal helper functions.
 */
static void 
init_enc_map(void)
{
	int             i;
	memset(enc_map, 0, sizeof(enc_map));
	for (i = 0; enc_set[i]; i++) {
		enc_map[(unsigned char) enc_set[i]] = i;
	}
	memset(r50_map, 255, sizeof(r50_map));
	for (i = 0; r50_set[i]; i++) {
		r50_map[(unsigned char) r50_set[i]] = i;
	}
	enc_map_ready = 1;
}
/*
 * Pack 12-character usernames into 4 16-bit words.
 */
static void 
decode_T_field(unsigned short text[4], unsigned char *s)
{
	int             i, j;
	unsigned short  a, ci[3];
	unsigned char  *start;

	start = s;
	for (i = 0; i < 4; i++) {
		a = text[i];
		ci[0] = a % 40;
		a = a / 40;
		ci[1] = a % 40;
		ci[2] = a / 40;

		for (j = 0; j < 3; j++)
			*s++ = r50_set[ci[j]];
	}
	while (s > start) {	/* trim trailing blanks */
		if (s[-1] != ' ')
			break;
		--s;
	}
	*s = '\0';
}
static int 
encode_T_field(unsigned char *s, unsigned short text[4])
{
	int             i, j, status;
	unsigned short  c, ci[3];

	status = 1;
	/*
         * Produce 4 words of output.
         */
	for (i = 0; i < 4; i++) {
		/*
		 * Map ASCII to a limited 40 character set to pack 3 characters in
		 * 16 bits (40^3 = 64000).
		 */
		for (j = 0; j < 3; j++) {
			unsigned short  cindex;
			if (*s)
				c = *s++;
			else
				c = ' ';
			cindex = r50_map[c];
			if (cindex > 40) {
				cindex = r50_map['.'];
				status = 0;	/* remember invalid character
						 * seen. */
			}
			ci[j] = cindex;	/* value is offset */
		}
		text[i] = ci[2] * 1600 + ci[1] * 40 + ci[0];
	}
	return status;
}
/*
 * Encode buffers into string using format string. "$V$%64q%8b%16w%2l"
 */
static int 
encode_6bit(void *buffer, size_t bufsize, const char *fmt,...)
{
	va_list         ap;
	int             fpos, apos, out_count, field_size, arg_bytes;
	uaf_lword       accum, zext;
	unsigned char  *out, *arg;
	/*
         * Outer loop, parse format argument
         */
	out_count = 0;
	out = buffer;
	accum = 0;		/* accumulate bits for output */
	apos = 0;		/* Number of bits accumulated */
	va_start(ap, fmt);
	for (fpos = 0; fmt[fpos]; fpos++)
		if (fmt[fpos] == '%') {
			/*
			 * Interpret format directive, either % or nnt when nn is number
			 * of bits and t is type: q-quadqord, l-longword, w-word, b-byte.
		      	 */
			fpos++;
			if (fmt[fpos] == '%') {
				out[out_count++] = '%';
				continue;
			}
			for (field_size = 0; isdigit(fmt[fpos]); fpos++) {
				field_size = field_size * 10 + (fmt[fpos] - '0');
			}
			arg = va_arg(ap, unsigned char *);

			switch (fmt[fpos]) {
			case 'T':	/* quadword of rad-50 'Text' */
			case 'q':
				arg_bytes = 8;
				break;
			case 'l':
				arg_bytes = 4;
				break;
			case 'w':
				arg_bytes = 2;
				break;
			case 'b':
				arg_bytes = 1;
				break;
			default:
				/* Invalid directive, ignored */
				arg_bytes = 0;
				break;
			}
			/*
			 * Input text string is rad-50 compressed quadword followed by
		         * expanded string if  'T' format.
			 */
			if (fmt[fpos] == 'T') {
				encode_T_field(arg + 8, (unsigned short *) arg);
			}
			/*
			 * Produce field_size bits of output from arg_bytes input bytes.
			 * Zero fill if argument size less than field size.
			 */
			while (field_size > 0) {
				if (arg_bytes > 0) {
					zext = *arg++;
					arg_bytes--;
				} else {
					zext = 0;
				}
				accum = accum | (zext << apos);
				apos += (field_size > 8) ? 8 : field_size;
				field_size -= 8;
				while (apos >= 6) {
					if ((out_count + 1) >= bufsize) {
						out[out_count] = 0;
						return out_count;
					}
					out[out_count++] = enc_set[accum & 63];
					accum = (accum >> 6);
					apos = apos - 6;
				}
			}
		} else {
			/*
			 * Move character verbatim from format string to output,  flush
			 * partial encoding bits accumulated.
			 */
			while (apos > 0) {
				if ((out_count + 1) >= bufsize) {
					out[out_count] = 0;
					return out_count;
				}
				out[out_count++] = enc_set[accum & 63];
				accum = accum >> 6;
				apos = (apos >= 6) ? apos - 6 : 0;
			}
			out[out_count++] = (unsigned char) fmt[fpos];
		}
	va_end(ap);

	out[out_count] = 0;	/* terminate string */
	return out_count;

}
/*
 * Decode encoded buffer via format string.
 */
static int 
decode_6bit(void *buffer, const char *fmt,...)
{
	va_list         ap;
	int             fpos, apos, out_count, field_size, arg_bytes;
	uaf_lword       accum;
	unsigned char  *in, *arg;
	/*
         * Outer loop, parse format argument
         */
	out_count = 0;
	in = buffer;
	accum = 0;		/* accumulate bits for output */
	apos = 0;		/* Number of bits accumulated */
	va_start(ap, fmt);
	for (fpos = 0; fmt[fpos]; fpos++)
		if (fmt[fpos] == '%') {
			/*
			 * Interpret format directive, either % or nnt when nn is number
			 * of bits and t is type: q-quadqord, l-longword, w-word, b-byte.
			 */
			fpos++;
			if (fmt[fpos] == '%') {
				if (*in != '%') {
					return -1;
				}
				in++;
				continue;
			}
			for (field_size = 0; isdigit(fmt[fpos]); fpos++) {
				field_size = field_size * 10 + (fmt[fpos] - '0');
			}
			arg = va_arg(ap, unsigned char *);

			switch (fmt[fpos]) {
			case 'T':	/* rad-50 encoding of text */
			case 'q':
				arg_bytes = 8;
				break;
			case 'l':
				arg_bytes = 4;
				break;
			case 'w':
				arg_bytes = 2;
				break;
			case 'b':
				arg_bytes = 1;
				break;
			default:
				/* Invalid directive, ignored */
				arg_bytes = 0;
				break;
			}
			/*
			 * Move field_size bits of input to argument buffer.
			 * Zero fill if field is less than argument size.
			 */
			while (field_size > 0) {
				while (apos < 8) {
					if (!*in)
						break;	/* buffer short */
					accum |= (enc_map[*in++] << apos);
					apos += 6;
				}
				if (arg_bytes > 0) {
					*arg++ = accum & 255;
					arg_bytes--;
					accum = (accum >> 8);
					apos = (apos < 8) ? 0 : apos - 8;
				}
				field_size -= 8;
			}
			if (arg_bytes > 0) {
				/*
				 * Flush accumulated bits and zero remaining
				 * buffer
				 */
				while (apos > 0) {
					*arg++ = accum & 255;
					arg_bytes--;
					accum = (accum >> 8);
					apos = (apos < 8) ? 0 : apos - 8;
				}
				while (arg_bytes > 0) {
					*arg++ = 0;
					arg_bytes--;
				}
			}
			/*
			 * Convert T quadword back to ASCII zero-terminated string.
			 */
			if (fmt[fpos] == 'T') {
				decode_T_field((unsigned short *) (arg - 8), arg);
			}
		} else {
			/*
			 * non-directive characters in format must match characters in
			 * input buffer.
			 */
			if (*in != fmt[fpos])
				return -1;
			in++;
		}
	va_end(ap);

	return 0;

}
/****************************************************************************/
/*
 * External functions.
 */
char           *
uaf_hash_encode(
	struct uaf_hash_info * pwd,	/* Input argument */
	char encoded[UAF_ENCODE_SIZE])
{				/* Output buffer */
	int             count;
	/*
         * Collapse the preserved flags to a 4-bit field.
         */
	pwd->opt = 0;
	if (pwd->flags & UAI$M_PWDMIX)
		pwd->opt |= 1;
	/*
         * Encode the string.
         */
	count = encode_6bit(encoded, UAF_ENCODE_SIZE, SYSUAF_FMT,
		&pwd->hash, &pwd->alg, &pwd->salt, &pwd->username, &pwd->opt);
	return encoded;
}

int 
uaf_hash_decode(
	char *encoded,		/* Input argument */
	struct uaf_hash_info * pwd)
{				/* Output buffer */
	int             count;
	/*
         * Decode the string.
         */
	count = decode_6bit(encoded, SYSUAF_FMT, &pwd->hash, &pwd->alg,
		&pwd->salt, &pwd->username, &pwd->opt);
	/*
         * expand 4-bit flags field into standard UAI$_FLAGS bits.
         */
	pwd->flags = ((pwd->opt & 1) ? UAI$M_PWDMIX : 0);

	return 1;
}

int 
uaf_getuai_info(
	const char *username,	/* Username to find */
	struct uaf_hash_info * pwd,	/* Password info from UAF record */
	struct uaf_hash_info * pwd2,	/* Secondary password info */
	struct uaf_account_info * acct)
{
	static long     uai_ctx = -1;	/* protected by uaf_static mutex */
#ifdef VMS
	$DESCRIPTOR(username_dx, "");
	char            owner[32];	/* counted string */
	char            defdev[32];	/* counted string */
	char            defdir[64];	/* counted string */
	char            defcli[32];
#define ITEM_LIST_SIZE 12
	struct {
		unsigned short  length, code;
		void           *buffer;
		unsigned short *retlen;
	}               item[ITEM_LIST_SIZE];
#define SET_ITEM(ITM,l,c,b,r) ITM.length=(l); ITM.code=(c); \
	ITM.buffer=(b); ITM.retlen=(r)
	unsigned short  retlen[ITEM_LIST_SIZE];
	int             status, i;
	/*
         * Build item list for GETUAI call, the acct argument is optional.
         */
	SET_ITEM(item[0], 8, UAI$_PWD, &pwd->hash, 0);
	SET_ITEM(item[1], 4, UAI$_FLAGS, &pwd->flags, 0);
	SET_ITEM(item[2], 2, UAI$_SALT, &pwd->salt, 0);
	SET_ITEM(item[3], 1, UAI$_ENCRYPT, &pwd->alg, 0);
	SET_ITEM(item[4], 8, UAI$_PWD2, &pwd2->hash, 0);
	SET_ITEM(item[5], 2, UAI$_ENCRYPT2, &pwd2->alg, 0);
	SET_ITEM(item[6], 0, 0, 0, 0);
	for (i = 0; (i < 12) && username[i]; i++)
		pwd->username.s[i] = toupper(username[i]);
	pwd->username.s[i] = '\0';
	memcpy(pwd2->username.s, pwd->username.s, sizeof(pwd->username.s));
	if (acct) {
		strcpy(acct->username, pwd->username.s);
		SET_ITEM(item[6], 4, UAI$_UIC, &acct->uic, 0);
		SET_ITEM(item[7], 32, UAI$_OWNER, owner, &retlen[7]);	/* counted str! */
		SET_ITEM(item[8], 32, UAI$_DEFDEV, defdev, &retlen[8]);
		SET_ITEM(item[9], 64, UAI$_DEFDIR, defdir, &retlen[9]);
		SET_ITEM(item[10], 32, UAI$_DEFCLI, defcli, &retlen[10]);
		SET_ITEM(item[11], 0, 0, 0, 0);
	}
	/*
         * Call system to get the information and fixup.  Serialize.
         */
	pthread_mutex_lock(&uaf_static);
	username_dx.dsc$a_pointer = (char *) username;
	username_dx.dsc$w_length = strlen(username);
	status = SYS$GETUAI(0, &uai_ctx, &username_dx, item, 0, 0, 0);
	pthread_mutex_unlock(&uaf_static);
	if ((status & 1) == 0) {
		return status;
	}
	pwd2->salt = pwd->salt;	/* Fill in pwd2 */
	pwd2->flags = pwd->flags;
#define CONVERT_COUNTED(zs,cs) memcpy(zs,&cs[1],cs[0]); zs[cs[0]] = '\0'
	if (acct) {
		CONVERT_COUNTED(acct->owner, owner);
		CONVERT_COUNTED(acct->home_dir, defdev);
		memcpy(&acct->home_dir[defdev[0]], &defdir[1], defdir[0]);
		acct->home_dir[defdev[0] + defdir[0]] = '\0';
		CONVERT_COUNTED(acct->shell, defcli);
	}
	return status;
#else
	if (uai_ctx == -1) {
#ifdef NOPTHREADS
		uai_ctx = -2;
		uaf_static = 1;
#else
		pthread_mutex_lock(&uaf_static);	/* just to fool compiler */
		uai_ctx = -2;
		pthread_mutex_unlock(&uaf_static);
#endif
	}
	return 0;
#endif
}
/*
 * Convert string to RAD-50 packed format and vice versa.  Conversion is
 * unpacked-to-packed if to_packed argument is non-zero and packed-to-unpacked
 * if zero.  Return value is 1 on success and 0 if string has invalid
 * characters for conversion.
 */
int 
uaf_packed_convert(struct uaf_packed_text * username, int to_packed)
{
	int             status;
	if (to_packed) {
		status = encode_T_field((unsigned char *) username->s, username->r40);
	} else {
		decode_T_field(username->r40, (unsigned char *) username->s);
		status = 1;
	}
	return status;
}
/*
 * Hash caller-supplied password and compare with hash value stored
 * in uaf_hash_info structure.  Return value is 1 for match.
 */
int 
uaf_test_password(
	struct uaf_hash_info * pwd,
	const char *password,	/* clear text password */
	int replace_if)
{				/* Update pwd if false */
	char            uc_username[32], uc_password[32];
	$DESCRIPTOR(username_dx, uc_username);
	$DESCRIPTOR(password_dx, "");
	int             status, i, ulen;
	uaf_qword       hashed_password;
	/*
         * Build VMS descriptors for system service arguments.
         */
	ulen = strlen(pwd->username.s);
	if (ulen > sizeof(uc_username) - 1)
		return 0;	/* name too long */
	strcpy(uc_username, pwd->username.s);
	username_dx.dsc$w_length = ulen;

	password_dx.dsc$w_length = strlen(password);
	if (pwd->flags & UAI$M_PWDMIX) {	/* take password verbatim */
		password_dx.dsc$a_pointer = (char *) password;
	} else {
		/*
		 * Upcase password.
		 */
		password_dx.dsc$a_pointer = uc_password;
		if (password_dx.dsc$w_length > sizeof(uc_password))
			password_dx.dsc$w_length = sizeof(uc_password);
		for (i = 0; i < password_dx.dsc$w_length; i++)
			uc_password[i] = toupper(password[i]);
	}
	/*
         * Try private implementation first (optimized) and fall back
         * to OS routine if failed.
         */
#ifndef NOLGI
	status = hash_password(&hashed_password,
		&password_dx, pwd->alg, pwd->salt, &username_dx);

	if (((status & 1) == 0)) {
		printf("Retry... (lgi$hpwd2 status %d)\n", status);
#else
	if (1) {
#endif
#ifdef VMS
		status = SYS$HASH_PASSWORD(&password_dx, pwd->alg, pwd->salt,
			&username_dx, &hashed_password);
#else
		printf("Error returned by lib$hpwd2\n");
		exit(2);
#endif
	}
	if (pwd->flags & UAI$M_PWDMIX)
		memset(uc_password, 0, sizeof(uc_password));

	if ((status & 1) == 0)
		return 0;

	if (UAF_QW_EQL(hashed_password, pwd->hash))
		return 1;

	if (replace_if)
		pwd->hash = hashed_password;

	return 0;
}
/*
 * Pull in hash function definition from an included file.  May be overriden
 * to allow alternate implentations.
 */
#define UAF_INCLUDED_FROM_ENCODE 1
/*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/
/*
 * Revised: 22-JUL-2009			Re-factor exponent to improve speed.
 * Revised: 14-AUG-2009			Re-locate 'P' initialization.
 * Revised: 25-AUG-2011			Misc. code cleanup.
 *
 */

/***************************************************************************/
/*

Title:		hash_password (originaly LGI$HPWD)
Author:		Shawn A. Clifford		(sysop@robot.nuceng.ufl.edu)
Date:		19-FEB-1993
Revised:	7-JUL-2009	Rework by David Jones for John the Ripper
Purpose:	Portable C version of the last 3 encryption methods for DEC's
		password hashing algorithms.

Usage:		status = lgi$hpwd(&out_buf, &uaf_buf, encrypt, salt, &unam_buf);

		int lgi$hpwd2(
		    unsigned long long *output_hash,
		    string *password,
		    unsigned char *encrypt,
		    unsigned short *salt,
		    string *username)

		Where:
			output_hash = 8 byte output buffer descriptor
			   password = n character password string descriptor
			    encrypt = 1 byte; value determines algorithm to use
				      0 -> CRC algorithm   (not supported)
				      1 -> Purdy algorithm
				      2 -> Purdy_V
				      3 -> Purdy_S (Hickory algorithm)
			       salt = 2 byte (word) random number
			   username = up to 31 character username descriptor

			status is either SS$_NORMAL (1) or SS$_ABORT (44)

Compilation:	VAXC compiler, or 'gcc -traditional -c' to get hpwd.o file
		on a Un*x machine.

Comments:	The overall speed of this routine is not great.  This is
		dictated by the performance of the EMULQ routine which emulates
		the VAX EMULQ instruction (see notes in the EMULQ routine).
		If anyone can improve performance, or finds bugs in this
		program, please send me e-mail at the above address.

*/

#ifdef VMS
#include <ssdef.h>
#else
#define SS$_ABORT	44
#define SS$_BADPARAM     20
#define SS$_NORMAL	1
#define __SSDEF_LOADED	1
#endif

typedef struct dsc$descriptor_s string;


/*
 *	Create a quadword data type as successive longwords.
 */
typedef union {
	uaf_lword       ulw[2];
	uaf_qword       uqw;
	char            b[8];
}               quad;
#define l_low ulw[0]		/* Low order longword in the quadword */
#define l_high ulw[1]		/* High order longword in the quadword */


/*
 *	The following table of coefficients is used by the Purdy polynmial
 *	algorithm.  They are prime, but the algorithm does not require this.
 */
static const struct {
	quad            C1, C2, C3, C4, C5;
}               C = {{{
			-83, -1
}
}, {{
		-179, -1
}
}, {{
		-257, -1
}
}, {{
		-323, -1
}
},
{{
		-363, -1
}
}
};



/** Function prototypes **/

static void     COLLAPSE_R2(string * r3, quad * r4, char r7);
/*
 * r3 -> input descriptor r4 -> output buffer r7 : Method 3 flag
 */
static void     Purdy(quad * U);/* U -> output buffer */
#ifdef NOLONGLONG
static void     PQMOD_R0(quad * U);	/* U MOD P       */
#endif
static void 
PQEXP_pair(quad * U, int higbits, uaf_lword n0,
	uaf_lword n1, quad * result0, quad * result1);
static void 
PQADD_R0(quad * U,
	quad * Y,
	quad * result);		/* U + Y MOD P   */
static void 
PQMUL_ADD(quad * U,
	quad * Y, quad * X,
	quad * result);		/* U * Y MOD P   */
static void 
EMULQ(uaf_lword a,
	uaf_lword b,
	quad * result);		/* (a * b)	 */
static void 
PQLSH_R0(quad * U,
	quad * result);		/* 2^32*U MOD P  */


/** RELATIVELY GLOBAL variables **/

static uaf_lword MAXINT = 4294967295;	/* Largest 32 bit number */
static uaf_lword a = 59;	/* 2^64 - 59 is the biggest quadword prime */
static quad     P;		/* P = 2^64 - 59 */

/*
 * Initialize P as 2^64 - a = MAXINT.MAXINT - 1 + 1  Assume called from
 * main thread.
 */
int 
uaf_init(void)
{
	static int      initialized = 0;
	if (!enc_map_ready)
		init_enc_map();
	if (initialized)
		return 0;
	/* since MAXINT.MAXINT = 2^64 - 1		 */
	P.l_high = MAXINT;
	P.l_low = MAXINT - a + 1;
	initialized = 1;
	return 1;
}

/** LGI$HPWD entry point **/

static int 
hash_password(
	uaf_qword * output_hash,
	string * password,
	unsigned char encrypt,
	unsigned short salt,
	string * username)
{
	string         *r3;	/* Holds descriptors for COLLAPSE_R2 */
	quad           *r4;	/* Address of the output buffer */
	uaf_lword       r5;	/* Length of username */
	char            r7 = 0,	/* Flag for encryption method # 3 */
	               *bytptr;	/* Pointer for adding in random salt */
	quad            qword;	/* Quadword form of the output buffer */
	char            uname[13];	/* buffer for padded username (PURDY) */

	/*
	 * --------------------------------------------------------------------
	 * ----
	 */


	/* Check for invalid parameters */
	if ((encrypt < 1) || (encrypt > 3)) {
		exit(SS$_BADPARAM);
	}
	if (username->dsc$w_length > 31) {
		printf("Internal coding error, username is more than 31 bytes long.\n");
		exit(SS$_ABORT);
	}
	/* Setup pointer references */
	r3 = password;		/* 1st COLLAPSE uses the password desc.   */
	r4 = &qword;		/* @r4..@r4+7 equals obuf */
	r5 = username->dsc$w_length;
	r7 = (encrypt == 3);

	/* Clear the output buffer (zero the quadword) */
	r4->ulw[0] = 0;
	r4->ulw[1] = 0;
	UAF_QW_SET(*output_hash, 0);

	/* Check for the null password and return zero as the hash value if so */
	if (password->dsc$w_length == 0) {
		return SS$_NORMAL;
	}
	switch (encrypt) {
		int             ulen;
	case UAI$C_AD_II:	/* CRC algorithm with Autodin II poly */
		/* As yet unsupported */
		return SS$_BADPARAM;

	case UAI$C_PURDY:	/* Purdy algorithm */

		/* Use a blank padded username */
		strncpy(uname, "            ", sizeof(uname));
		strncpy(uname, username->dsc$a_pointer, r5);
		username->dsc$a_pointer = (char *) &uname;
		username->dsc$w_length = 12;
		break;

	case UAI$C_PURDY_V:	/* Purdy with blanks stripped */
	case UAI$C_PURDY_S:	/* Hickory algorithm; Purdy_V with rotation */

		/*
		 * Check padding.  Don't count blanks in the string length.
		 * Remember:  r6->username_descriptor   the first word is
		 * length, then 2 bytes of class information (4 bytes total),
		 * then the address of the buffer.  Usernames can not be
		 * longer than 31 characters.
		 */
		for (ulen = username->dsc$w_length; ulen > 0; ulen--) {
			if (username->dsc$a_pointer[ulen - 1] != ' ')
				break;
			username->dsc$w_length--;
		}

		/* If Purdy_S:  Bytes 0-1 => plaintext length */
		if (r7) {
			r4->ulw[0] = password->dsc$w_length;
		}
		break;
	}


	/* Collapse the password to a quadword U; buffer pointed to by r4 */
	COLLAPSE_R2(r3, r4, r7);
	/* r3 already points to password descriptor */


	/* Add random salt into the middle of U */
	/* This has to be done byte-wise because the Sun will not allow you */
	/*
	 * to add unaligned words, or it will give you a bus error and core
	 * dump :)
	 */
	bytptr = &r4->b[3 + 1];
	*bytptr += (char) (salt >> 8);	/* Add the high byte */

	/* Check for carry out of the low byte */
	bytptr--;
	if ((short) ((unsigned char) *bytptr + (unsigned char) (salt & 0xff)) > 255) {
		*(bytptr + 1) += 1;	/* Account for the carry */
	}
	*bytptr += (char) (salt & 0xff);	/* Add the low byte */



	/* Collapse the username into the quadword */
	r3 = username;		/* Point r3 to the valid username desriptor */
	COLLAPSE_R2(r3, r4, r7);

	/* U (qword) contains the 8 character output buffer in quadword format */

	/* Run U through the polynomial mod P */
	Purdy(r4);

	/* Write qword (*r4) back into the output buffer */
	*((quad *) output_hash) = qword;

	/* Normal exit */
	return SS$_NORMAL;

}				/* LGI$HPWD */




/***************	Functions Section	*******************/


/***************
   COLLAPSE_R2
 ***************/

static void 
COLLAPSE_R2(string * r3, quad * r4, char r7)
/*
 *    r3 :  input string descriptor
 *    r4 :  output buffer (quadword)
 *    r7 :  flag (1) if using Hickory method (encrypt = 3)
 *
 * This routine takes a string of bytes (the descriptor for which is pointed
 * to by r3) and collapses them into a quadword (pointed to by r4).  It does
 * this by cycling around the bytes of the output buffer adding in the bytes of
 * the input string.  Additionally, after every 8 characters, each longword in
 * the resultant hash is rotated by one bit (PURDY_S only).
 *
 */

{
	unsigned short  r0, r1;
	uaf_lword       rotate;
	char           *r2, mask = -8;

	/*
	 * --------------------------------------------------------------------
	 * -
	 */

	r0 = r3->dsc$w_length;	/* Obtain the number of input bytes */

	if (r0 == 0)
		return;		/* Do nothing with empty string */

	r2 = r3->dsc$a_pointer;	/* Obtain pointer to input string */

	for (; (r0 != 0); r0--) {	/* Loop until input string exhausted */

		r1 = (~mask & r0);	/* Obtain cyclic index into out buff */
		r4->b[r1] += *r2++;	/* Add in this character */

		if ((r7) && (r1 == 7)) {	/* If Purdy_S and last byte
						 * ... */
			/* Rotate first longword one bit */
			rotate = r4->ulw[0] & 0x80000000;
			r4->ulw[0] <<= 1;
			rotate >>= 31;
			r4->ulw[0] |= rotate;

			/* Rotate second longword one bit */
			rotate = r4->ulw[1] & 0x80000000;
			r4->ulw[1] <<= 1;
			rotate >>= 31;
			r4->ulw[1] |= rotate;
		}		/* if Purdy_S */
	}			/* for loop */

	return;

}				/* COLLAPSE_R2 */


/************
   PQADD_R0
 ************/

static void 
PQADD_R0(quad * U, quad * Y, quad * result)
/*
 * U, Y : quadwords that we want to add
 *
 *
 * Computes the sum U + Y MOD P where P is of the form P = 2^64 - a.
 * U, Y are quadwords less than P.
 *
 * Fixed with the help of the code written by Terence Lee (DEC/HKO).
 */

{
	uaf_lword       carry = 0;
#ifndef NOLONGLONG
	static const unsigned long long maxqw = 0xffffffffffffffffLL;
	static const unsigned long long modulus = (0xffffffffffffffffLL - 59 + 1);

	if ((maxqw - U->uqw) < Y->uqw)
		carry = 1;
	result->uqw = U->uqw + Y->uqw;
	if (!carry) {
		if (result->uqw < modulus)
			return;
		result->uqw -= (modulus);
	} else {
		UAF_QW_ADD(result->uqw, a);	/* missing case of uqw > max-a */
	}
#else				/* Compiler doesn't support long long */


	/* Add the low longwords, checking for carry out */
	if ((MAXINT - U->l_low) < Y->l_low)
		carry = 1;
	result->l_low = U->l_low + Y->l_low;

	/* Add the high longwords */
	result->l_high = U->l_high + Y->l_high + carry;

	/* Test for carry out of high bit in the quadword */
	if ((MAXINT - U->l_high) < (Y->l_high + carry))
		carry = 1;
	else
		carry = 0;

	/* Check if we have to MOD P the result */
	if (!carry && Y->l_high != MAXINT)
		return;		/* Outta here? */
	if (Y->l_low > (MAXINT - a))
		carry = 1;
	else
		carry = 0;
	result->l_low += a;	/* U + Y MOD P */
	result->l_high += carry;
#endif

	return;			/* Outta here! */
}				/* PQADD_R0 */


/*********
   EMULQ
 *********/

static void 
EMULQ(uaf_lword a, uaf_lword b, quad * result)
/*
 * a, b   : longwords that we want to multiply (quadword result)
 * result : the quadword result
 *
 * This routine knows how to multiply two unsigned longwords, returning the
 * unsigned quadword product.
 *
 * Originally I wrote this using a much faster routine based on a
 * divide-and-conquer strategy put forth in Knuth's "The Art of Computer
 * Programming, Vol. 2, Seminumerical Algorithms" but I could not make the routine
 * reliable.  There is some sort of fixup for sign compensation, much like for the
 * VAX EMULQ instruction (where for each signed argument you must add the other
 * argument to the high longword).
 *
 * If anyone can improve this routine, please send me source code.
 *
 * The original idea behind this algorithm was to build a 2n-by-n matrix for the
 * n-bit arguments to fill, shifting over each row like you would do by hand.  I
 * found a simple way to account for the carries and then get the final result,
 * but the routine was about 4 to 5 times slower than it is now.  Then I realized
 * that if I made the variables global, that would save a little time on
 * allocating space for the vectors and matrices.  Last, I removed the matrix, and
 * added all the values to the 'temp' vector on the fly.  This greatly increased
 * the speed, but still nowhere near Knuth or calling LIB$EMULQ.
 */

{
#ifndef NOLONGLONG
	uaf_qword       qw_a, qw_b;

	qw_a = a;		/* promote to long long */
	qw_b = b;
	result->uqw = qw_a * qw_b;
#else
	char            bin_a[32], bin_b[32];
	char            temp[64], i, j;
	char            retbuf[8];


	/* Initialize */
	for (i = 0; i <= 63; i++)
		temp[i] = 0;
	for (i = 0; i <= 31; i++)
		bin_a[i] = bin_b[i] = 0;
	for (i = 0; i <= 7; retbuf[i] = 0, i++);


	/* Store the binary representation of a & b */
	for (i = 0; i <= 31; i++) {
		bin_a[i] = a & 1;
		bin_b[i] = b & 1;
		a >>= 1;
		b >>= 1;
	}


	/* Add in the shifted multiplicand */
	for (i = 0; i <= 31; i++) {

		/*
		 * For each 1 in bin_b, add in bin_a, starting in the ith
		 * position
		 */
		if (bin_b[i] == 1) {
			for (j = i; j <= i + 31; j++)
				temp[j] += bin_a[j - i];
		}
	}


	/* Carry into the next position and set the binary value */
	for (j = 0; j <= 62; j++) {
		temp[j + 1] += temp[j] / 2;
		temp[j] = temp[j] % 2;
	}
	temp[63] = temp[63] % 2;


	/* Convert binary bytes back into 8 packed bytes. */
	/* LEAST SIGNIFICANT BYTE FIRST!!!  This is LITTLE ENDIAN format. */
	for (i = 0; i <= 7; i++) {
		for (j = 0; j <= 7; j++)
			retbuf[i] += temp[i * 8 + j] * (1 << j);
	}


	/* Copy the 8 byte buffer into result */
	memcpy((char *) result, retbuf, 8);
#endif

	return;

}				/* EMULQ */



/************
   PQMOD_R0
 ************/
#ifdef NOLONGLONG
static void 
PQMOD_R0(quad * U)
/*
 * U : output buffer (quadword)
 *
 * This routine replaces the quadword U with U MOD P, where P is of the form
 * P = 2^64 - a			(RELATIVELY GLOBAL a = 59)
 *   = FFFFFFFF.FFFFFFFF - 3B + 1	(MAXINT = FFFFFFFF = 4,294,967,295)
 *   = FFFFFFFF.FFFFFFC5
 *
 * Method:  Since P is very nearly the maximum integer you can specify in a
 * quadword (ie. P = FFFFFFFFFFFFFFC5, there will be only 58 choices for
 * U that are larger than P (ie. U MOD P > 0).  So we check the high longword in
 * the quadword and see if all its bits are set (-1).  If not, then U can not
 * possibly be larger than P, and U MOD P = U.  If U is larger than MAXINT - 59,
 * then U MOD P is the differential between (MAXINT - 59) and U, else
 * U MOD P = U.  If U equals P, then U MOD P = 0 = (P + 59).
 */

{

	/* Check if U is larger/equal to P.  If not, then leave U alone. */

	if (U->l_high == P.l_high && U->l_low >= P.l_low) {
		U->l_low += a;	/* This will have a carry out, and ...	 */
		U->l_high = 0;	/* the carry will make l_high go to 0   */
	}
	return;

}				/* PQMOD_R0 */
#else				/* Replace function with macro */
#define PQMOD_R0(x) if ( (x)->uqw < P.uqw ); else (x)->uqw += a;
#endif


static void 
PQEXP_pair(quad * U, int highbit, uaf_lword n0,
	uaf_lword n1, quad * result0, quad * result1)
/*
 * U        : pointer to output buffer (quadword)
 * highbit  : Highest bit set in n0/n1 +1
 * n0,n1    : unsigned longword (exponent for U)
 *
 * The routine returns U^n MOD P where P is of the form P = 2^64-a.
 * U is a quadword, n is an unsigned longword, P is a RELATIVELY GLOBAL quad.
 * We optimize operation by generating two results for a single U re-using
 * the powers of 2 vector.
 *
 * The method comes from Knuth, "The Art of Computer Programing, Vol. 2", section
 * 4.6.3, "Evaluation of Powers."  This algorithm is for calculating U^n with
 * fewer than (n-1) multiplies.  The result is U^n MOD P only because the
 * multiplication routine is MOD P.  Knuth's example is from Pingala's Hindu
 * algorthim in the Chandah-sutra.
 */

{
	quad            b2[32], *b2ptr;	/* sucessive squaring of base.  */
	int             is_one;	/* True if  intermediate value== 1 */
	/*
         * Build series where b2[i] = U^(2^i) through repeated squaring.
         */
	b2[0] = *U;
	for (b2ptr = b2; highbit > 1; highbit--) {
		PQMUL_ADD(b2ptr, b2ptr, 0, b2ptr + 1);	/* b2[i] = U ^ (2^i) */
		b2ptr++;
	}
	/*
         * Compute U^n0 by multiply the factors corresponding to set bits in the
         * the exponent (U^(a+b) = U^a * U^b).  Assume exponent non-zero.
         */
	is_one = 1;		/* Skip initializing result, set flag instead. */
	for (b2ptr = b2; n0 != 0; b2ptr++) {
		if (n0 % 2) {	/* Test low bit */
			if (is_one) {	/* Use assign instead of multiply */
				is_one = 0;
				*result0 = *b2ptr;
			} else
				PQMUL_ADD(result0, b2ptr, 0, result0);
		}
		n0 /= 2;	/* Shift right to test next bit next round */
	}
	/*
         * Compute U^n1, but only if supplied by caller.
         */
	if (!result1)
		return;
	is_one = 1;
	for (b2ptr = b2; n1 != 0; b2ptr++) {
		if (n1 % 2) {
			if (is_one) {
				is_one = 0;
				*result1 = *b2ptr;
			} else
				PQMUL_ADD(result1, b2ptr, 0, result1);
		}
		n1 /= 2;
	}
	return;

}				/* PQEXP_pair */



/************
   PQMUL_ADD
 ************/

static void 
PQMUL_ADD(quad * U, quad * Y, quad * X, quad * result)
/*
 * U, Y, A : Input values, A optional.
 * result  : Output, result = (U*Y)+A
 *
 * Computes the product U*Y+X MOD P where P is of the form P = 2^64 - a.
 * U, Y, X are quadwords less than P.  The result is returned in result.
 *
 * The product may be formed as the sum of four longword multiplications
 * which are scaled by powers of 2^32 by evaluating:
 *
 *	2^64*v*z + 2^32*(v*y + u*z) + u*y
 *      ^^^^^^^^   ^^^^^^^^^^^^^^^^   ^^^
 *      part1       part2 & part3     part4
 *
 * The result is computed such that division by the modulus P is avoided.
 *
 * u is the low longword of  U;	u = U.l_low
 * v is the high longword of U;	v = U.l_high
 * y is the low longword of  Y;	y = Y.l_low
 * z is the high longword of Y;	z = Y.l_high
 */

{
	quad            stack, part1, part2, part3;

	EMULQ(U->l_high, Y->l_high, &stack);	/* Multiply   v*z      */

	PQMOD_R0(&stack);	/* Get   (v*z) MOD P   */


	/*** 1st term ***/
	PQLSH_R0(&stack, &part1);	/* Get   2^32*(v*z) MOD P  */


	EMULQ(U->l_high, Y->l_low, &stack);	/* Multiply   v*y      */

	PQMOD_R0(&stack);	/* Get   (v*y) MOD P   */

	EMULQ(U->l_low, Y->l_high, &part2);	/* Multiply   u*z      */

	PQMOD_R0(&part2);	/* Get   (u*z) MOD P   */

	PQADD_R0(&stack, &part2, &part3);	/* Get   (v*y + u*z)   */

	PQADD_R0(&part1, &part3, &stack);	/* Get   2^32*(v*z) + (v*y +
						 * u*z) */


	/*** 1st & 2nd terms ***/
	PQLSH_R0(&stack, &part1);	/* Get   2^64*(v*z) + 2^32*(v*y + u*z) */


	EMULQ(U->l_low, Y->l_low, &stack);	/* Multiply   u*y      */


	/*** Last term ***/
	PQMOD_R0(&stack);

	if (X)
		PQADD_R0(&stack, X, &stack);

	PQADD_R0(&part1, &stack, result);	/* Whole thing */


	return;

}				/* PQMUL_R2 */



/************
   PQLSH_R0
 ************/

static void 
PQLSH_R0(quad * U, quad * result)
/*
 * Computes the product 2^32*U MOD P where P is of the form P = 2^64 - a.
 * U is a quadword less than P.
 *
 * This routine is used by PQMUL in the formation of quadword products in
 * such a way as to avoid division by modulus the P.
 * The product 2^64*v + 2^32*u is congruent a*v + 2^32*U MOD P.
 *
 * u is the low longword in U
 * v is the high longword in U
 */

{
	quad            stack;


	/* Get	a*v   */
	EMULQ(U->l_high, a, &stack);

	/* Form  Y = 2^32*u  */
	U->l_high = U->l_low;
	U->l_low = 0;

	/* Get	U + Y MOD P  */
	PQADD_R0(U, &stack, result);

	return;

}				/* PQLSH_R0 */


/*********
   Purdy
 *********/

static void 
Purdy(quad * U)
/*
 * U : input/output buffer (quadword)
 *
 * This routine computes  f(U) = p(U) MOD P.  Where P is a prime of the form
 * P = 2^64 - a.  The function p is the following polynomial:
 *
 * 		X^n0 + X^n1*C1 + X^3*C2 + X^2*C3 + X*C4 + C5
 *                ^^^^^^^^^^^^^^   ^^^^^^^^^^^^^^^^^^^^^^^^^^^
 *		    part1		  part2
 *
 * Note:  Part1 is calculated by its congruence  (X^(n0-n1) + C1)*X^n1
 *        finding X^n1, X^(n0-n1), X^(n0-n1)+C1, then
 *        (X^(n0-n1) + C1)*X^n1  which equals  X^n0 + X^n1*C1
 *
 *  To minimize the number of multiplications, we evaluate
 *
 *  f(U) =  ((U^(n0-n1) + C1)*U^(n1-1)*U + (U*C2 + C3)*U + C4)*U + C5
 *          ^^^^^^^^^^^^^^^^^^^^^^^^^^     ^^^^^^^^^^^^^^^^^^^^^^^^^^
 *                    part1                         part2
 *
 * The number of multiplies needed for an exponentiation is equal to
 * the bit position of the most signficant bit + the number of other set
 * bits in the binary expansion of the exponent.  By treating U^n1 as
 * U*U^(n1-1), we can replace n1-1 with 448*37449 to compute the
 * value in (1+15+5+8+2)=31 multiplies rather than (23+18)=41 multiplies used
 * for a direct U^n1.  Note that because of the pqexp_pair optimization, the
 * U^(n0-n1) calculation is done with just 3 multiplies.
 *
 */

{
	quad           *Cptr = (quad *) & C;	/* Address of table (C) */
	static const uaf_lword n0 = 16777213,	/* These exponents are prime,
						 * but this is  */
	                n1 = 16777153;	/* not required by the algorithm 	   */
	static const uaf_lword na = 37449, nb = 448;	/* U^n1 = (U^na)^nb * U  */
	/* n0 = 2^24 - 3;  n1 = 2^24 - 63;  n0-n1 = 60 */
	quad            X,	/* The variable X in the polynomial */
	                X_n1, X_n0Mn1, X_na, X_na_nb,	/* X raised to na, na*nb */
	                part1,	/* Collect the polynomial ... */
	                part2;	/* ... in two parts */


	/*
	 * --------------------------------------------------------------------
	 * -
	 */


	/* Ensure U less than P by taking U MOD P  */
	/* Save copy of result:  X = U MOD P       */
	X.l_low = U->l_low;
	X.l_high = U->l_high;

	if (UAF_QW_GEQ(X.uqw, P.uqw))
		UAF_QW_ADD(X.uqw, a);	/* Take X mod P    */

	PQEXP_pair(&X, 16, (n0 - n1), na, &X_n0Mn1, &X_na);
	PQADD_R0(&X_n0Mn1, Cptr, &part1);	/* X^(n1-n0) + C1   */

	PQEXP_pair(&X_na, 9, nb, 0, &X_na_nb, 0);	/* X^na^nb 	    */
	PQMUL_ADD(&X_na_nb, &X, 0, &X_n1);	/* X^na^nb*X = X^n1 */

	/* Part 1 complete except for multiply by X^n1 */

	Cptr++;			/* Point to C2		 */

	PQMUL_ADD(&X, Cptr, Cptr + 1, &part2);	/* part2= X*C2 + C3	 */

	Cptr += 2;		/* C4				 */
	PQMUL_ADD(&X, &part2, Cptr, &part2);	/* part2 = part2*X + C4 */
	/* (X^2*C2 + X*C3 + C4)	 */

	Cptr++;			/* C5			 */
	PQMUL_ADD(&X, &part2, Cptr, &part2);	/* part2=part2*X + C5       	 */
	/* X^3*C2 + X^2*C3 + X*C4 + C5	 */
	/* Part 2 complete */


	/* Do final multiply for part1 and add in part2 to get final result.  */

	PQMUL_ADD(&part1, &X_n1, &part2, (quad *) U);

	/* Outta here... */
	return;

}				/* Purdy */
